"use strict";
/**
 * Copyright (c) 2019-2022 Red Hat, Inc.
 * This program and the accompanying materials are made
 * available under the terms of the Eclipse Public License 2.0
 * which is available at https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * Contributors:
 *   Red Hat, Inc. - initial API and implementation
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.isCommandExists = exports.isCheFlavor = exports.isPartOfEclipseChe = exports.newListr = exports.getImageNameAndTag = exports.addTrailingSlash = exports.getEmbeddedTemplatesDirectory = exports.safeLoadFromYamlFile = exports.getProjectVersion = exports.getProjectName = exports.newError = exports.sleep = exports.base64Decode = void 0;
const tslib_1 = require("tslib");
const fs = require("fs-extra");
const os = require("node:os");
const yaml = require("js-yaml");
const path = require("node:path");
const context_1 = require("../context");
const Listr = require("listr");
const flags_1 = require("../flags");
const eclipse_che_1 = require("../tasks/installers/eclipse-che/eclipse-che");
const constants_1 = require("../constants");
const commandExists = require("command-exists");
const execa = require("execa");
const pkjson = require('../../package.json');
function base64Decode(arg) {
    return Buffer.from(arg, 'base64').toString('ascii');
}
exports.base64Decode = base64Decode;
function sleep(ms) {
    return new Promise(resolve => {
        setTimeout(resolve, ms);
    });
}
exports.sleep = sleep;
function newError(message, cause) {
    const error = new Error(message);
    error.cause = cause;
    error.stack += `\nCause: ${cause.stack}`;
    return error;
}
exports.newError = newError;
function getProjectName() {
    return pkjson.name;
}
exports.getProjectName = getProjectName;
function getProjectVersion() {
    return pkjson.version;
}
exports.getProjectVersion = getProjectVersion;
function safeLoadFromYamlFile(filePath) {
    return yaml.load(fs.readFileSync(filePath).toString());
}
exports.safeLoadFromYamlFile = safeLoadFromYamlFile;
function getEmbeddedTemplatesDirectory() {
    // Embedded templates are located in the templates directory that is in the project/installation root:
    // chectl
    //  |- templates
    //  |- src
    //  |   |- utls.ts
    //  |  ...
    //  |- lib
    //  |   |- util.js
    // ... ...
    // __dirname is
    //   project_root/src if dev mode,
    //   installation_root/lib if run from an installed location
    return path.join(__dirname, '..', '..', 'templates');
}
exports.getEmbeddedTemplatesDirectory = getEmbeddedTemplatesDirectory;
function addTrailingSlash(url) {
    if (url.endsWith('/')) {
        return url;
    }
    return url + '/';
}
exports.addTrailingSlash = addTrailingSlash;
function getImageNameAndTag(image) {
    let imageName;
    let imageTag;
    if (image.includes('@')) {
        // Image is referenced via a digest
        const index = image.indexOf('@');
        imageName = image.slice(0, Math.max(0, index));
        imageTag = image.slice(Math.max(0, index + 1));
    }
    else {
        // Image is referenced via a tag
        const lastColonIndex = image.lastIndexOf(':');
        if (lastColonIndex === -1) {
            // Image name without a tag
            imageName = image;
            imageTag = 'latest';
        }
        else {
            const beforeLastColon = image.slice(0, Math.max(0, lastColonIndex));
            const afterLastColon = image.slice(Math.max(0, lastColonIndex + 1));
            if (afterLastColon.includes('/')) {
                // The colon is for registry port and not for a tag
                imageName = image;
                imageTag = 'latest';
            }
            else {
                // The colon separates image name from the tag
                imageName = beforeLastColon;
                imageTag = afterLastColon;
            }
        }
    }
    return [imageName, imageTag];
}
exports.getImageNameAndTag = getImageNameAndTag;
function newListr(tasks, collapse = false) {
    const flags = context_1.CheCtlContext.getFlags();
    const options = { renderer: flags[flags_1.LISTR_RENDERER_FLAG], collapse };
    return new Listr(tasks, options);
}
exports.newListr = newListr;
function isPartOfEclipseChe(resource) {
    var _a, _b;
    return ((_b = (_a = resource === null || resource === void 0 ? void 0 : resource.metadata) === null || _a === void 0 ? void 0 : _a.labels) === null || _b === void 0 ? void 0 : _b['app.kubernetes.io/part-of']) === 'che.eclipse.org';
}
exports.isPartOfEclipseChe = isPartOfEclipseChe;
function isCheFlavor() {
    return eclipse_che_1.EclipseChe.CHE_FLAVOR === constants_1.CHE;
}
exports.isCheFlavor = isCheFlavor;
function isCommandExists(commandName) {
    return tslib_1.__awaiter(this, void 0, void 0, function* () {
        if (commandExists.sync(commandName)) {
            return true;
        }
        // commandExists.sync fails if not executable command exists in the same directory.
        // Double check without accessing local file.
        // The check above there is for backward compatability.
        const whereCommand = os.platform() === 'win32' ? 'where' : 'whereis';
        try {
            yield execa(whereCommand, [commandName]);
            return true;
        }
        catch (_a) { }
        return false;
    });
}
exports.isCommandExists = isCommandExists;
//# sourceMappingURL=utls.js.map